import {Observable, Subscription} from 'rxjs';
import {ICommand, ICommandAdapter, IResolveCommandModelEvent} from './command.abstraction';
import {
  IAdvanceQueryAdapter,
  IAggregate,
  IFilter,
  IGroup,
  IQueryCriteria,
  IQueryExecutionGroupResult,
  IQueryExecutionResult,
  ISort
} from './query.abstraction';

export interface IDataSource<TQuery extends IQueryCriteria, TModel>
{
  resolveCommandModelByName<TCommand extends ICommand>(event: IResolveCommandModelEvent<TModel>) : Observable<TCommand>;
  executeCommandByName<TCommand extends ICommand, TResult>(name: string, command: TCommand) : Observable<TResult>;
  query(query: TQuery): Subscription;
  executeQuery(query: TQuery): Observable<IQueryExecutionGroupResult<TModel>>;
  refresh(): Subscription;
  resolveIdField(model: TModel) : TModel[keyof TModel];
  clear(): void;
  updateData(value: IQueryExecutionResult<TModel> & IQueryExecutionGroupResult<TModel>): void;
  replaceDataWithArray(items: TModel[]): void;
  replaceDataWithSingle(item: TModel): void;

  data$: Observable<IQueryExecutionResult<TModel> & IQueryExecutionGroupResult<TModel> | null>;
  loading$: Observable<boolean>;
  validationError$: Observable<IDataSourceValidationError>;
  notifyMessage$: Observable<IDataSourceNotifyMessage>;
  commandStarted$: Observable<IDataSourceCommandStarted>;

  data?: IQueryExecutionResult<TModel> & IQueryExecutionGroupResult<TModel>;

  sorts: ISort[];
  filters: IFilter[];
  groups: IGroup[];
  aggregates: IAggregate[];
  pageSize: number;
  page: number;
}

export interface IDataSourceOptions<TQuery extends IQueryCriteria, TResult>{
  transport: IDataSourceTransportOptions<TQuery, TResult>;
  idField?: keyof TResult;
  resolveIdField?: (model: TResult) => TResult[keyof TResult];
  defaultCriteria: IQueryCriteria;
}

export interface IDataSourceTransportOptions<TQuery extends IQueryCriteria, TResult> {
  query: IDataSourceQueryAdapterOptions<TQuery, TResult>;
  commands: {
    [name: string]: IDataSourceCommandAdapterOptions<TResult>;
  };
}

export interface IDataSourceQueryAdapterOptions<TQuery extends IQueryCriteria, TResult> {
  adapter: IAdvanceQueryAdapter<TQuery, TResult>;
}

export interface IDataSourceCommandAdapterOptions<TModel> {
  adapter: ICommandAdapter<ICommand, any>;
  resolveCommandModel?: (event: IResolveCommandModelEvent<TModel>) => Observable<any>;
}

export interface IDataSourceCommandStarted {
  name: string;
  command: ICommand;
}

export interface IDataSourceNotifyMessage {
  type: 'warning' | 'info' | 'success' | 'error';
  message: string;
  messageParams?: any;
}


export interface IDataSourceError {
  type: 'message' | 'validation';
}

export interface IDataSourceErrorMessage extends IDataSourceError {
  message: string;
}

export interface IDataSourceValidationError extends IDataSourceError
{
  errors: { [field: string]: string[]; }
}
